#!/usr/bin/perl -w

package ScControld;

# TODO: all instance variables should be gotten by calling getter methods.

=head1 NAME

ScControld - RPC::XML::Server subclass to controll sc_ldap2dnsdhcpd.pl;

=head1 Synopsis

  use ScControld;
  my $sc_controld = ScControld->new();
  $sc_controld->setup_methods();
  $sc_controld->server_loop();

in C<RPC::XML::Methods> we can get instance information for this class by using the following 
code fragment

  sub example_xmlrpc_method {
    my $self = shift;
    my ($other_params,...) = @_;
    
    my $function_queue = $self->{'sc_controld'}->{'function_queue'};
    ... 
  }

=head1 Description

This Serverclass is used to control various aspects of the ScCache in sc_ldap2dnsdhcpd.pl.

=head2 RPC::XML::Methods which are implemented in this server

=cut

use strict;
use warnings;

# we are a subclass of RPC::XML::Server
use base qw(RPC::XML::Server);
our $VERSION = "1.0";

use Data::Dumper;
use Log::Log4perl;
use Time::Local;
use Time::HiRes qw( usleep );
use Storable;
use Cache::FastMmap;
use Net::LDAP;
use RPC::XML::Server;
use RPC::XML::Procedure;
use Net::Server::MultiType;

use SmartClient::Config;
use SmartClient::Constant;
use SmartClient::Util;
use SmartClient::LdapUtil;
use SmartClient::Queue;
use SmartClient::SharedFunction;
use SmartClient::RpcXmlUtil;

# SmartClient::IP is needed for dumpEntry, since objects of that type can be transferred through the shared function queue
use SmartClient::IP;

use constant NOT_FOUND_IN_CACHE => 1;
use constant MARKED_AS_DIRTY    => 2;

our $version = '1.0';

=head2 new

Constructor for this class

Special care has to be taken after calling C<< $class->SUPER >>, since it returns a newly created hash reference
as a blessed object. So we are inserting our instance variables into that hash under the key C<sc_controld>.

=cut

sub new
{
    my $class = shift;
    if ( ref $class ) {
        $class = ref $class;
    }

    # TODO: is this really needed?
    # force auto flush of output buffer
    $|++;

    my $config = SmartClient::Config->new(
        '/etc/smartclient/base',     '/etc/smartclient/controld',
        '/etc/smartclient/ldap2dns', '/etc/smartclient/leases2ldap',
        '/etc/smartclient/ldap2dnsdhcpd',
    );

    # check if configuration attributes are defined
    $config->require(
        'ALL_CONTROLD_URL',                         'CONTROLD_LOG4PERL_CONF',
        'LEASES2LDAP_MMAP_FILE_DISLOCATED_DEVICES', 'CONTROLD_HA_MODE'
    );
    my $ha_mode                    = $config->get('CONTROLD_HA_MODE');
    my $dislocated_device_filename =
      $config->{'LEASES2LDAP_MMAP_FILE_DISLOCATED_DEVICES'};

    my $device_status_mmap_filename =
      $config->{'LDAP2DNS_DEVICE_STATUS_MMAP_FILE'};

    my $function_queue_page_size = $config->get('SHARED_FUNCTIONS_MMAP_PAGE_SIZE', 1)
      || '512k';
    my $function_queue_num_pages = $config->get('SHARED_FUNCTIONS_MMAP_NUM_PAGES', 1)
      || '313';

    my $function_queue = SmartClient::SharedFunction->new(
        'shm_id_queue'     => 'l2dq',
        'mmap_result_file' => '/var/cache/smartclient/shared_function_results',
        'mmap_page_size'   => $function_queue_page_size,
        'mmap_num_pages'   => $function_queue_num_pages,
        ##'init_mmap_result_file' => 0
    );

    my $xml_rpc_default_timeout = $config->get('CONTROLD_XML_RPC_TIMEOUT')
      || 30;

    Log::Log4perl::init( $config->get('CONTROLD_LOG4PERL_CONF') );
    my $logger = Log::Log4perl->get_logger();

    my $params = {
        'port' => getUrlPort( $config->get('ALL_CONTROLD_URL') ),
        @_
    };
    my $port = $params->{'port'};

    my $data = {
        'config'                      => $config,
        'default_timeout'             => $xml_rpc_default_timeout,
        'function_queue'              => $function_queue,
        'ha_mode'                     => $ha_mode,
        'device_status_mmap_filename' => $device_status_mmap_filename,
        'dislocated_device_filename'  => $dislocated_device_filename
    };

    # configure access restrictions
    my $allow_hosts = $config->get( 'CONTROLD_XML_RPC_ALLOW_HOSTS', 1 );
    if ( defined $allow_hosts ) {
        $data->{'allow_hosts'} = qr/^$allow_hosts$/o;
    }

    my $deny_hosts = $config->get( 'CONTROLD_XML_RPC_DENY_HOSTS', 1 );
    if ( defined $deny_hosts ) {
        $data->{'deny_hosts'} = qr/^$deny_hosts$/o;
    }

    # now call new on SUPER class, which should be an RPC::XML::Server
    my $super_data = $class->SUPER::new(
        'port'        => $port,
        'no_http'     => 1,
        'sc_controld' => $data
    );
    return $super_data;
}

=head2 allow_deny_hook

implementing the access hook of L<Net::Server>. It will check the keys C<< $self->{sc_controld}->{deny_hosts} >>
and C<< $self->{sc_controld}->{allow_hosts} >> to grant or deny access to clients.

=over 4

=item Parameter

C<Net::Server> reference to self 

=item Returns

boolean: true if access is granted, false otherwise

=back

=cut

sub allow_deny_hook
{
    my $self   = shift;
    my $logger = Log::Log4perl->get_logger();

    # check wether access control is enabled
    if (   $self->{'sc_controld'}->{'allow_hosts'}
        or $self->{'sc_controld'}->{'deny_hosts'} )
    {
        my $client_addr   = $self->{'server'}->{'peeraddr'};
        my $deny_hosts_re = $self->{'sc_controld'}->{'deny_hosts'};
        $logger->debug( "check access for " . $client_addr );

        # check if this client is explicitly disallowed
        if ( defined $deny_hosts_re and $client_addr =~ $deny_hosts_re ) {
            $logger->debug( "access denied for " . $client_addr );
            return 0;
        }
        my $allow_hosts_re = $self->{'sc_controld'}->{'allow_hosts'};

        # the client has to be allowed explicitly
        if ( defined $allow_hosts_re and $client_addr =~ $allow_hosts_re ) {
            $logger->debug( "access granted for " . $client_addr );
            return 1;
        }

        # the client is not allowed, or denied, but that means for us deny access
        return 0;
    }

    # no access checks were configured, so always allow access.
    return 1;
}

=head2 getDislocatedDeviceFilename

=over 4

=item Parameter

=item Return

=back

=cut

sub getDislocatedDeviceFilename
{
    my $self = shift;
    return $self->{'sc_controld'}->{'dislocated_device_filename'};
}

# TODO: document method and add error checking

=head2 getDeviceStatusMmapFilename

=over 4

=item Parameter

=item Return

=back

=cut

sub getDeviceStatusMmapFilename
{
    my $self = shift;
    return $self->{'sc_controld'}->{'device_status_mmap_filename'};

}

# TODO: document method and add error checking

=head2 getDefaultTimeout

=over 4

=item Parameter

=item Return

=back

=cut

sub getDefaultTimeout
{
    my $self = shift;
    return $self->{'sc_controld'}->{'default_timeout'};

}

# TODO: document method and add error checking

=head2 getConfig

=over 4

=item Parameter

=item Return

=back

=cut

sub getConfig
{
    my $self = shift;
    return $self->{'sc_controld'}->{'config'};
}

=head2 setHaMode

=over 4

=item Parameter

=item Return

=back

=cut

sub setHaMode
{
    my $self = shift;
    my $mode = shift;

    $self->{'sc_controld'}->{'ha_mode'} = $mode;
    return $self->{'sc_controld'}->{'ha_mode'};
}

=head2 getHaMode

=over 4

=item Parameter

=item Return

=back

=cut

sub getHaMode
{
    my $self = shift;
    return $self->{'sc_controld'}->{'ha_mode'};
}

# TODO: document method

=head2 getFunctionQueue

=over 4

=item Parameter

=item Return

=back

=cut

sub getFunctionQueue
{
    my $self = shift;

    # TODO: more error checking
    return $self->{'sc_controld'}->{'function_queue'};
}

=head2 xmlrpc_getRandom

generates a random number. Mainly used to test the server easily.

=over 4

=item Parameter

none

=item Returns

C<RPC::XML::double> random value

=back

=cut

sub xmlrpc_getRandom
{

    #sleep 5;
    return RPC::XML::double->new( rand(10) );
}

=head2 xmlrpc_setPrimary

is intended to switch this server into primary mode for ha.
B<This will do nothing at the moment>

=over 4

=item Parameter

none

=item Return

C<RPC::XML::string> containing the actual ha_mode

=back

=cut

sub xmlrpc_setPrimary
{
    my $self = shift;
    $self->setHaMode('ha_mode_primary');
    setup_methods();
    return RPC::XML::string->new( $self->getHaMode() );
}

=head2 xmlrpc_setSecondary

is intended to switch this server into primary mode for ha.
B<This will do nothing at the moment>

=over 4

=item Parameter

none

=item Return

C<RPC::XML::string> containing the actual ha_mode

=back

=cut

sub xmlrpc_setSecondary
{
    my $self = shift;
    $self->setHaMode('ha_mode_secondary');
    setup_methods();
    return RPC::XML::string->new( $self->getHaMode() );
}

=head2 xmlrpc_getHaMode

get the configured ha mode from the server.

=over 4

=item Parameter

none

=item Return

C<RPC::XML::string> containing the actual ha_mode

=back

=cut

sub xmlrpc_getHaMode
{
    my $self = shift;
    return RPC::XML::string->new( $self->getHaMode() );

}

=head2 xmlrpc_createConfig

asks the sc_ldap2dhcpdns.pl script to write all configurations for
dhcp and dns.

=over 4

=item Parameter

none

=item Return

C<RPC::XML::struct> containing

=over 4

=item C<code>

C<RPC::XML::int> XMLRPC_OK

=item C<state>

state from function queue

=back

=back

=cut

sub xmlrpc_createConfig
{
    my $self = shift;

    my $logger = Log::Log4perl->get_logger();

    $logger->info("XMLRPC Method createConfig was called");

    my $function_queue          = $self->getFunctionQueue();
    my $xml_rpc_default_timeout = $self->getDefaultTimeout;
    my $id          = $function_queue->enqueue('main::createConfig');
    my $return_href =
      $function_queue->getResult( $id, $xml_rpc_default_timeout );

    #return RPC::XML::string->new( $return_href->{'state'} );
    return RPC::XML::struct->new(
        'code'  => RPC::XML::int->new(XMLRPC_OK),
        'state' => RPC::XML::string->new( $return_href->{'state'} )
    );
}

=head2 xmlrpc_changeDeviceType

=over 4

=item Parameter

=item Return

=back

=cut

sub xmlrpc_changeDeviceType
{
    my $self               = shift;
    my $href_xmlrpc_params = shift;

    my $function_queue = $self->getFunctionQueue();
    my $id             =
      $function_queue->enqueue( 'main::callChangeDeviceType',
        %{$href_xmlrpc_params} );
    return $self->doGetResult(
        'id'           => $id,
        'errormessage' => "callChangeDeviceType failed: timeout"
    );
}

=head2 doGetResult

internally used method to get results from the function queue
and pack them as C<RPC::XML>

=over 4

=item hash containing

=over 4

=item id

id of the function call in the queue

=item timeout

timeout in seconds to wait for a result

=item errormessage

message to display in case of an error

=back

=back

=cut

sub doGetResult
{
    my $self   = shift;
    my $logger = Log::Log4perl->get_logger();

    my %params = (
        'id'           => undef,
        'timeout'      => $self->getDefaultTimeout(),
        'errormessage' => "failed (timeout)",
        @_
    );

    my ( $package, $filename, $line, $subroutine ) = caller(1);
    my $function_queue = $self->getFunctionQueue();

    my $id = $params{'id'};

    my $code    = 0;
    my $message = "";
    my %response;

    my $return_href = $function_queue->getResult( $id, $params{'timeout'} );

    if ( !$return_href ) {

        $code    = XMLRPC_SHARED_FUNCTION_TIMEOUT;
        $message =
          "shared function call for $subroutine [id: $id]: "
          . $params{'errormessage'};

        $logger->error($message);

    } else {
        %response = %{$return_href};
        $logger->debug( "$subroutine [id: $id] shared function result: "
              . Dumper(%response) );
    }

    return RPC::XML::struct->new(
        'code'    => RPC::XML::int->new($code),
        'message' => RPC::XML::string->new($message),
        %response,
        'id' => RPC::XML::string->new($id),
    );
}

=head2 xmlrpc_getResult

checks answer for a delayed shared function call

=over 4

=item Parameter

hash containing

=over 4

=item id

id of the call in the function queue

=item timeout

timeout in seconds to wait before abort

=back

=item Return

passes return value from doGetResult through

=back

=cut

sub xmlrpc_getResult
{
    my $logger = Log::Log4perl->get_logger();

    my $self               = shift;
    my $href_xmlrpc_params = shift;
    my %params             = (
        'id'      => undef,
        'timeout' => $self->getDefaultTimeout(),
        %{$href_xmlrpc_params}
    );
    return $self->doGetResult(%params);
}

=head2 xmlrpc_notifyOfModification

notifies ScCache in sc_ldap2dnsdhcpd.pl of changes in a given ldap subtree 

=over 4

=item Parameter

array containing a list of changed dn's as the local roots for the changes

=item Return

passes through the result of doGetResult

=back

=cut

sub xmlrpc_notifyOfModification
{
    my $self               = shift;
    my $aref_xmlrpc_params = shift;

    my $function_queue = $self->getFunctionQueue();
    my $id = $function_queue->enqueue( 'main::callNotifyOfModification',
        $aref_xmlrpc_params );
    return $self->doGetResult( 'id' => $id );
}

=head2 xmlrpc_notifyOfDelete

=over 4

=item Parameter

=item Return

=back

=cut

sub xmlrpc_notifyOfDelete
{
    my $self = shift;

    # TODO: why is this an array and not a reference to an array?
    my @list_of_dns = @_;

    my $function_queue = $self->getFunctionQueue();
    my $logger         = Log::Log4perl->get_logger();
    $logger->debug( Dumper(@list_of_dns) );

    my $id = $function_queue->enqueue( 'main::notifyOfDelete', \@list_of_dns );
    return $self->doGetResult( 'id' => $id );
}

=head2 getDeviceStatus

used internally to get status for a given dn from ScCache

There are two possiple ways to get that information.

=over 4

=item direct access

This will get device status information by placing a call in
the function queue

=item acces through mmap cache

This method is fast, since it will not go through function queue. ScCache will
have to update the mmap cache for this to work correctly.

=back

=over 4

=item Parameter

device dn

timeout in seconds (optional)

=item Return

passes through hash reference from doGetDeviceStatus

=back

=cut

sub doGetDeviceStatus()
{
    my $self   = shift;
    my $logger = Log::Log4perl->get_logger();

    my $device_dn = ldap_normalize(shift);
    my $timeout   = shift;
    if ( !defined($timeout) ) {
        $timeout = $self->getDefaultTimeout();
    }

    my $href_result;
    my $device_status_mmap_filename = $self->getDeviceStatusMmapFilename();

    # there are 2 methods:
    # using the mmap file, generated by ldap2dnsdhcp
    # or calling a shared function on ldap2dnsdhcp.
    # if mmap is available (config file) we use this method
    if ($device_status_mmap_filename) {

        my $config = $self->getConfig();
        my $num_pages =
          $config->get( 'DEVICE_STATUS_CACHE_MMAP_NUM_PAGES', 1 )
          || '137';
        my $page_size =
          $config->get( 'DEVICE_STATUS_CACHE_MMAP_PAGE_SIZE', 1 )
          || '64k';

        # Mmap is initialised by the server,
        # therefore it is required,
        # that the client reconnects from time to time.
        # Otherwise no change is detected,
        # if server is restarted
        my $device_status = Cache::FastMmap->new(
            'share_file' => $device_status_mmap_filename,
            'init_file'  => 0,
            ## for the value see comments in SmartClient::ScCache::createDeviceStatusFile
            'num_pages' => $num_pages,
            'page_size' => $page_size,
        );
        $href_result = $device_status->get($device_dn);

        # TODO: to_be_changed should not be visible here, could we use another
        # instance of DeviceStatusCache? That one should be with init_file == 0 of course.
        if ( ref $href_result and $href_result->{'to_be_changed'} ) {
            $href_result->{'code'}    = MARKED_AS_DIRTY;
            $href_result->{'message'} = "$device_dn is marked as to be changed";
        }

        if ( !$href_result ) {
            $href_result->{'code'}    = NOT_FOUND_IN_CACHE;
            $href_result->{'message'} = "$device_dn not found in cache";
        }

    } else {
        my $function_queue = $self->getFunctionQueue();
        my $id             =
          $function_queue->enqueue( 'main::callGetDeviceStatus', $device_dn );
        $href_result = $function_queue->getResult( $id, $timeout );

    }

    return $href_result;
}

=head2 xmlrcp_getDeviceStatus

=over 4

=item Parameter

hash containing

=over 4

=item C<device_dn> dn of the device for which to get device status

=back

=item Returns

C<RPC::XML::struct> containing

=over 4

=item code

C<RPC::XML::int> status of the call

=item message

C<RPC::XML::string> short description of the call

=item C<%message> gotten from doGetDeviceStatus is weaved in

=back

=back

=cut

sub xmlrpc_getDeviceStatus
{
    my $logger = Log::Log4perl->get_logger();

    my $self               = shift;
    my $href_xmlrpc_params = shift;
    my %params             = ( %{$href_xmlrpc_params} );
    my %response;

    my $device_dn = $params{'device_dn'};

    my $return_href = $self->doGetDeviceStatus($device_dn);

    my $code    = 0;
    my $message = "";

    if ($return_href) {

        # for now assume the hash consists only of string keys and
        # string values.
        foreach my $key ( keys %{$return_href} ) {
            $response{$key} = RPC::XML::string->new( $return_href->{$key} );
        }

        # netboot is defined to be boolean
        if ( $return_href->{'netboot'} ) {
            $response{'netboot'} =
              RPC::XML::boolean->new( $return_href->{'netboot'} );
        }
    } else {

        # this can only happen, if shared function is used
        # or mmap cache for this dn is marked as dirty
        $code    = XMLRPC_SHARED_FUNCTION_TIMEOUT;
        $message = "shared function timeout: no response for $device_dn";
        $logger->error($message);
    }

    $logger->debug( "$device_dn: " . "code: $code; " . "$message" );

    return (
        RPC::XML::struct->new(
            'code'    => RPC::XML::int->new($code),
            'message' => RPC::XML::string->new($message),
            %response
        )
    );
}

=head2 xmlrpc_waitForDeviceStatus

wait for specific settings for device_dn.
Can wait for netboot or dbkImageVersion 
or both of them 
(finished, if one of them has the required value)

=over 4

=item Parameter

    hash
        'device_dn'       => undef,
        'netboot'         => undef,
        'dbkImageVersion' => undef,
        'timeout'         => undef,


=item Return

    hash
        'code'            
        'message'         
        'device_dn'       
        'netboot'         
        'dbkImageVersion' 

=back

=cut

sub xmlrpc_waitForDeviceStatus
{
    my $self               = shift;
    my $href_xmlrpc_params = shift;
    my %params             = (
        'device_dn'       => undef,
        'netboot'         => undef,
        'dbkImageVersion' => undef,
        'timeout'         => undef,
        %{$href_xmlrpc_params}
    );

    my $logger    = Log::Log4perl->get_logger();
    my $device_dn = $params{'device_dn'};

    my $timeout    = $params{'timeout'};
    my $start_time = time();
    my $finish_time;
    if ( defined($timeout) ) {
        $finish_time = $start_time + $timeout;
    }

    my $required_netboot = $params{'netboot'};
    my $result_netboot;

    my $required_image_version = $params{'dbkImageVersion'};
    my $result_image_version;

    my $code    = 0;
    my $message = "";

    my $current_time = time();

    my $sleep     = 0;
    my $max_sleep = 10;

  LOOKUP_STATUS:
    while ( $code == 0 ) {

        my $href_result =
          $self->doGetDeviceStatus( $device_dn,
            ( $timeout || $self->getDefaultTimeout() ) );

        if ( ref $href_result ) {
            if ( $href_result->{'code'} == MARKED_AS_DIRTY ) {
                $href_result = undef;
            } else {
                $result_netboot       = $href_result->{'netboot'};
                $result_image_version = $href_result->{'dbkImageVersion'};
                $code                 = $href_result->{'code'} || 0;
                $message              = $href_result->{'message'} || "";
            }
        }

        if (   definedAndEquals( $required_netboot, $result_netboot )
            or
            definedAndEquals( $required_image_version, $result_image_version ) )
        {
            last LOOKUP_STATUS;
        }

        $current_time = time();
        if ( defined($timeout) and ( $current_time > $finish_time ) ) {

            #out of time, so give up
            $code += XMLRPC_TIMEOUT;
            last LOOKUP_STATUS;
        }

        # sleep a bit, to prevent using all CPU ressources
        $sleep += ( $sleep < $max_sleep );
        sleep $sleep;
    }

    $logger->debug( "$device_dn: code: $code; "
          . "netboot: "
          . ( $result_netboot   || "0" ) . ', '
          . ( $required_netboot || '-' ) . '; '
          . "image version: "
          . ( $result_image_version   || "0" ) . ", "
          . ( $required_image_version || "-" )
          . "; timeout: "
          . ( $timeout || "" ) )
      . "; duration: "
      . ( $current_time - $start_time );

    return RPC::XML::struct->new(
        'code'            => RPC::XML::int->new($code),
        'message'         => RPC::XML::string->new($message),
        'device_dn'       => RPC::XML::string->new($device_dn),
        'netboot'         => RPC::XML::boolean->new($result_netboot),
        'dbkImageVersion' => RPC::XML::string->new($result_image_version),
    );
}

# returns true if both arguments are undefined, or both are equal
sub definedAndEquals
{
    my ( $dit, $dat ) = @_;
    if ( defined $dit ) {

        # the other one has to be defined, too and must be equal to first argument
        return ( defined $dat and ( $dit eq $dat ) );
    } else {

        # dit was not defined, so we will always shout false
        return 0;
    }
}

=head2 xmlrpc_getScheduleUpdateWorkstations

calls callGetScheduleUpdateWorkstations in sc_ldap2dnsdhcpd.pl through function queue.

=over 4

=item Parameter

hash containing of

=over 4

=item C<location_dn>

=item C<refimage_dn>

=item C<devicetype_dn>

=back

=item Returns

passes through result from doGetResult

=back

=cut

sub xmlrpc_getScheduleUpdateWorkstations
{
    my $self               = shift;
    my $href_xmlrpc_params = shift;
    my %params             = (
        'location_dn'   => undef,
        'refimage_dn'   => undef,
        'devicetype_dn' => undef,
        %{$href_xmlrpc_params}
    );

    my $logger = Log::Log4perl->get_logger();

    #print STDERR "xmlrpc_getScheduleUpdateWorkstations\n";

    my $function_queue = $self->getFunctionQueue();
    my $id             =
      $function_queue->enqueue( 'main::callGetScheduleUpdateWorkstations',
        %params );

    #$logger->debug( "id: $id" );
    #print STDERR "id: $id\n";
    my $href_result =
      $function_queue->getResult( $id, $self->getDefaultTimeout() );

    return $self->doGetResult(
        'id'           => $id,
        'errormessage' => "getScheduleUpdateWorkstations failed: timeout"
    );
}

=head2 xmlrpc_setUpdate

places a call to callSetUpdate into function queue

=over 4

=item Parameter

hash containing of

=over 4

=item C<location_dn>

=item C<device_type_dn>

=item C<workstations>

=item C<tftpboot>

=item C<rsync>

=back

=item Returns 

passes result through from doGetResult

=back

=cut

sub xmlrpc_setUpdate
{
    my $logger = Log::Log4perl->get_logger();

    my $self               = shift;
    my $href_xmlrpc_params = shift;
    my %params             = (
        'location_dn'    => undef,
        'device_type_dn' => undef,
        'workstations'   => undef,
        'tftpboot'       => undef,
        'rsync'          => undef,
        %{$href_xmlrpc_params}
    );

    $logger->debug(
        'location_dn: ' . $params{'location_dn'},
        '; device_type_dn: ' . $params{'device_type_dn'},
        '; workstations: ' . $params{'workstations'},
        '; tftpboot: ' . $params{'tftpboot'},
        '; rsync: ' . $params{'rsync'}
    );

    my $function_queue = $self->getFunctionQueue();
    my $id = $function_queue->enqueue( 'main::callSetUpdate', %params );
    return $self->doGetResult(
        'id'           => $id,
        'errormessage' => "setUpdate failed: timeout"
    );
}

=head2 xmlrpc_dumpEntry

places a call to callDumpEntry into function queue

=over 4

=item Parameter

hash containing

=over 4

=item C<dn>

name of the entry to be inspected

=back

=item Returns

C<RPC::XML::struct> containing

=over 4

=item C<code>

status code of the call

=item C<message>

short description of the status of the call

=item C<id>

id of the call in the function queue

=item C<hash>

result from the call gotten from function queue as smart encoded C<RPC::XML::struct>

=back

=back

=cut

sub xmlrpc_dumpEntry
{
    my $self               = shift;
    my $href_xmlrpc_params = shift;
    my %params             = ( %{$href_xmlrpc_params} );

    my $response;
    my $dn                      = $params{'dn'};
    my $logger                  = Log::Log4perl->get_logger();
    my $function_queue          = $self->getFunctionQueue();
    my $xml_rpc_default_timeout = $self->getDefaultTimeout();
    my $id          = $function_queue->enqueue( 'main::callDumpEntry', $dn );
    my $return_href =
      $function_queue->getResult( $id, $xml_rpc_default_timeout );

    my $code    = 0;
    my $message = "";

    if ( ref $return_href ) {
        $message = "OK";

        #RPC::XML::smart_encode fails for objects in hashes (or something similar)
        $response = xmlrpc_convertRefToRpcXmlStructure($return_href);
    } else {
        $code    = XMLRPC_SHARED_FUNCTION_TIMEOUT;
        $message = "shared function timeout: no response for $dn";
        $logger->error($message);
    }

    return (
        RPC::XML::struct->new(
            'code'    => RPC::XML::int->new($code),
            'message' => RPC::XML::string->new($message),
            'id'      => RPC::XML::string->new($id),
            'hash'    => $response
        )
    );
}

# leases2ldap
# TODO: document the method

=head2 getDislocatedDevices

=over 4

=item Parameter

=item Returns

=back

=cut

sub getDislocatedDevices()
{
    my $self    = shift;
    my $actions = shift || ACTION_UNDEF;

    # Mmap is initialised by the server,
    # therefore it is required,
    # that the client reconnects from time to time.
    # Otherwise no change is detected,
    # if server is restarted
    #Log::Log4perl::init_once("/etc/smartclient/log4perl.noninteractive");
    #my $logger = Log::Log4perl->get_logger();
    my $dislocated_device_filename = $self->getDislocatedDeviceFilename();
    my $dislocated_devices         = Cache::FastMmap->new(
        share_file => $dislocated_device_filename,
        init_file  => 0,
    );
    my @list_ids    = ();
    my %result_hash = ();
    foreach my $id ( $dislocated_devices->get_keys(0) ) {

        my $dislocated_device = $dislocated_devices->get($id);

        # only return actions > ACTION_NOTHING
        # action: add, move, manual
        if ( $dislocated_device->{'action'} > $actions ) {
            push @list_ids, "$id";

            #$logger->debug( $mac );
            #print STDERR $mac;
            #$logger->debug( Dumper( $dislocated_device ) );
            my $href_xmlrpc_dislocated_device = {};
            xmlrpc_convertHashStrings(
                $href_xmlrpc_dislocated_device,
                $dislocated_device,
                [
                    'mac_address',     'lease_time_start',
                    'lease_ip',        'current_ldap_dn',
                    'location_dn_new', 'hostname',
                    'lease_time_end'
                ]
            );
            xmlrpc_convertHashInts(
                $href_xmlrpc_dislocated_device,
                $dislocated_device,
                [ 'action', 'permission_level', 'permission', 'lease_status' ]
            );

            #print STDERR Dumper($href_xmlrpc_dislocated_device);

            $result_hash{$id} =
              RPC::XML::struct->new($href_xmlrpc_dislocated_device);
        } else {

            # debug
            push @list_ids, "($id: " . $dislocated_device->{'action'} . ")";
        }
    }

    $self->log( 1,
        "$actions: getDislocatedDevices: " . join( ", ", @list_ids ) );

    return RPC::XML::struct->new(%result_hash);
}

=head2 xmlrpc_getDislocatedDevices

=over 4

=item Parameter

none

=item Returns

passes through result from C<getDislocatedDevices(ACTION_NOTHING)>

=back

=cut

sub xmlrpc_getDislocatedDevices
{

    my $self = shift;

    # return all dislocated devices with action > ACTION_NOTHING
    return $self->getDislocatedDevices(ACTION_NOTHING);
}

=head2 xmlrpc_getAllDislocatedDevices

=over 4

=item Parameter

none

=item Returns

passes through result from C<getDislocatedDevices()>

=back

=cut

sub xmlrpc_getAllDislocatedDevices
{
    my $self = shift;

    # return all dislocated devices
    # (with action > ACTION_UNDEF )
    return $self->getDislocatedDevices();
}

# TODO: document method

=head2 setDenyPermission

internally used method

=over 4

=item Parameters

=item Returns

=back

=cut

sub setDenyPermission
{
    my $self                   = shift;
    my $href_dislocated_device = shift;
    if ( $href_dislocated_device->{'mac_address'} ) {
        $href_dislocated_device->{'permission'} = 0;
    }
    return $href_dislocated_device;
}

# TODO: document method

=head2 xmlrpc_denyPermission

=over 4

=item Parameters

=item Returns

=back

=cut

sub xmlrpc_denyPermission
{
    my $self               = shift;
    my $href_xmlrpc_params = shift;

    my $id = $href_xmlrpc_params->{'id'};

    # Mmap is initialed by the server,
    # therefore it is required,
    # that the client reconnects from time to time.
    # Otherwise no change is detected,
    # if server is restarted
    #Log::Log4perl::init_once("/etc/smartclient/log4perl.noninteractive");
    #my $logger = Log::Log4perl->get_logger();
    my $dislocated_device_filename = $self->getDislocatedDeviceFilename();
    my $dislocated_devices         = Cache::FastMmap->new(
        share_file => $dislocated_device_filename,
        init_file  => 0,
    );

    my $result = $dislocated_devices->get_and_set( $id, \&setDenyPermission );

    #print STDERR $id . ":   " .Dumper( $result );

    my $code;
    my $message;

    # check if mac_address exist,
    # otherwise entry is assumed empty
    if ( $result->{mac_address} ) {
        $code    = RPC::XML::int->new(0);
        $message = RPC::XML::string->new("denyPermission( $id ) succeded");
    } else {
        $code    = RPC::XML::int->new(1);
        $message =
          RPC::XML::string->new(
            "denyPermission( $id ) failed: id $id does not exist");
    }
    return RPC::XML::struct->new( 'code' => $code, 'message' => $message );
}

# TODO: document methods

=head2 xmlrpc_relocateDislocatedDevice

=over 4

=item Parameters

=item Returns

=back

=cut

sub xmlrpc_relocateDislocatedDevice
{
    my $self               = shift;
    my $href_xmlrpc_params = shift;

    my $function_queue = $self->getFunctionQueue();
    my $id = $function_queue->enqueue( 'main::relocateDislocatedDevice',
        $href_xmlrpc_params );
    return $self->doGetResult(
        'id'           => $id,
        'errormessage' => "relocateDislocatedDevice failed: timeout"
    );
}

=head2 xmlrpc_getLeasesForDevice

=over 4

=item Parameters

=item Returns

=back

=cut

sub xmlrpc_getLeasesForDevice
{
    my $self               = shift;
    my $href_xmlrpc_params = shift;

    my $function_queue = $self->getFunctionQueue();
    my $id = $function_queue->enqueue( 'main::callGetLeasesForDevice',
        $href_xmlrpc_params->{'mac_addr'} );
    return $self->doGetResult(
        'id'           => $id,
        'errormessage' => "callGetLeasesForDevice failed: timeout"
    );
}

=head2 xmlrpc_deleteLeasesForDevice

=over 4

=item Parameters

=item Returns

=back

=cut

sub xmlrpc_deleteLeasesForDevice
{
    my $self               = shift;
    my $href_xmlrpc_params = shift;

    my $function_queue = $self->getFunctionQueue();
    my $id = $function_queue->enqueue( 'main::callDeleteLeasesForDevice',
        $href_xmlrpc_params->{'mac_addr'} );
    return $self->doGetResult(
        'id'           => $id,
        'errormessage' => "callDeleteLeasesForDevice failed: timeout"
    );
}

# end XML::RPC exported Methods

=head2 setup_methods

This method will add all methods in the local hash reference C<$methods> to the C<RPC::XML::Server>. 

Methods can be skipped if they are only useful in primary or secondary ha mode.

=head3 Example

this method has a big hash called C<$methods> which is a reference to a hash.
In this hash are all methods as key which shall be exported by this server. 
A simple example is given below:

  $ha_mode = 'ha_mode_secondary';
  $methods = {
    'notifyOfDelete' => {
        'ha_mode_primary' => 1,
        'ha_mode_secondary' => 1,
        'params' => {'name' => ... }
    },
    'setPrimary' => {'ha_mode' => HA_MODE_SECONDARY, 'params' => {'name' => ... }},
    'setSecondary' => {'ha_mode' => HA_MODE_PRIMARY, 'params' => {'name' => ... }},
  };

=over 4

=item Parameter

none

=item Return

reference to C<$self>

=back

=cut

sub setup_methods
{
    my $self = shift;

    # hash with all xmlrpc methods
    my $methods = {
        'getRandom' => {
            'ha_mode_primary'   => 1,
            'ha_mode_secondary' => 1,
            'params'            => {
                name      => 'getRandom',
                version   => $version,
                help      => 'get random number',
                code      => \&xmlrpc_getRandom,
                signature => ['int']
            }
        },
        'getResult' => {
            'ha_mode_primary' => 1,
            'params'          => {
                name      => 'getResult',
                version   => $version,
                help      => 'get pending result (for known id)',
                code      => \&xmlrpc_getResult,
                signature => ['struct struct']
            }
        },
        'getScheduleUpdateWorkstations' => {
            'ha_mode_primary' => 1,
            'params'          => {
                name    => 'getScheduleUpdateWorkstations',
                version => $version,
                help    =>
                  'get all workstations which are to be updated and of a given type',
                code      => \&xmlrpc_getScheduleUpdateWorkstations,
                signature => ['struct struct']
            }
        },
        'setUpdate' => {
            'ha_mode_primary' => 1,
            'params'          => {
                name      => 'setUpdate',
                version   => $version,
                help      => 'set device to update',
                code      => \&xmlrpc_setUpdate,
                signature => ['struct struct']
            }
        },
        'dumpEntry' => {
            'ha_mode_primary'   => 1,
            'ha_mode_secondary' => 1,
            'params'            => {
                name      => 'dumpEntry',
                version   => $version,
                help      => 'dump entry information into log file',
                code      => \&xmlrpc_dumpEntry,
                signature => ['struct struct']
            }
        },
        'getDeviceStatus' => {
            'ha_mode_primary' => 1,
            'params'          => {
                name      => 'getDeviceStatus',
                version   => $version,
                help      => 'get device status',
                code      => \&xmlrpc_getDeviceStatus,
                signature => ['struct struct']
            }
        },
        'waitForDeviceStatus' => {
            'ha_mode_primary'   => 1,
            'ha_mode_secondary' => 1,
            'params'            => {
                name      => 'waitForDeviceStatus',
                version   => $version,
                help      => 'wait for device state',
                code      => \&xmlrpc_waitForDeviceStatus,
                signature => ['struct struct']
            }
        },
        'getDislocatedDevices' => {
            'ha_mode_primary' => 1,
            'params'          => {
                name      => 'getDislocatedDevices',
                version   => $version,
                help      => 'get dislocated devices',
                code      => \&xmlrpc_getDislocatedDevices,
                signature => ['struct']
            }
        },
        'getAllDislocatedDevices' => {
            'ha_mode_primary' => 1,
            'params'          => {
                name      => 'getAllDislocatedDevices',
                version   => $version,
                help      => 'get all dislocated devices',
                code      => \&xmlrpc_getAllDislocatedDevices,
                signature => ['struct']
            }
        },
        'denyPermission' => {
            'ha_mode_primary' => 1,
            'params'          => {
                name      => 'denyPermission',
                version   => $version,
                help      => 'deny relocate permission for dislocated devices',
                code      => \&xmlrpc_denyPermission,
                signature => ['struct struct']
            }
        },
        'relocateDislocatedDevice' => {
            'ha_mode_primary' => 1,
            'params'          => {
                name    => 'relocateDislocatedDevice',
                version => $version,
                help    =>
                  'relocate dislocated devices in LDAP to its physical location',
                code      => \&xmlrpc_relocateDislocatedDevice,
                signature => ['struct struct']
            }
        },
        'notifyOfModification' => {
            'ha_mode_primary'   => 1,
            'ha_mode_secondary' => 1,
            'params'            => {
                name      => 'notifyOfModification',
                version   => $version,
                help      => 'notify sc_cache of modified dn',
                code      => \&xmlrpc_notifyOfModification,
                signature => ['struct array']
            }
        },
        'notifyOfDelete' => {
            'ha_mode_primary'   => 1,
            'ha_mode_secondary' => 1,
            'params'            => {
                name      => 'notifyOfDelete',
                version   => $version,
                help      => 'notify sc_cache of delete dn',
                code      => \&xmlrpc_notifyOfDelete,
                signature => ['struct array']
            }
        },
        'setPrimary' => {
            'ha_mode_secondary' => 1,
            'params'            => {
                name      => 'setPrimary',
                version   => $version,
                help      => 'set this node to primary ha mode',
                code      => \&xmlrpc_setPrimary,
                signature => ['string']
            }
        },
        'setSecondary' => {
            'ha_mode_primary' => 1,
            'params'          => {
                name      => 'setSecondary',
                version   => $version,
                help      => 'set this node to secondary ha mode',
                code      => \&xmlrpc_setSecondary,
                signature => ['string']
            }
        },
        'getHaMode' => {
            'ha_mode_primary'   => 1,
            'ha_mode_secondary' => 1,
            'params'            => {
                name      => 'getHaMode',
                version   => $version,
                help      => 'get ha mode',
                code      => \&xmlrpc_getHaMode,
                signature => ['string']
            }
        },
        'createConfig' => {
            'ha_mode_secondary' => 1,
            'params'            => {
                name      => 'createConfig',
                version   => $version,
                help      => 'trigger secondary host to create config files',
                code      => \&xmlrpc_createConfig,
                signature => ['struct']
            }
        },
        'changeDeviceType' => {
            'ha_mode_primary' => 1,
            'params'          => {
                name      => 'changeDeviceType',
                version   => $version,
                help      => '',
                code      => \&xmlrpc_changeDeviceType,
                signature => ['struct struct']
            }
        },
        'getLeasesForDevice' => {
            'ha_mode_primary' => 1,
            'params' => {
                name      => 'getLeasesForDevice',
                version   => $version,
                help      => 'get all lease entries for a mac address',
                code      => \&xmlrpc_getLeasesForDevice,
                signature => ['struct struct']
            }
        },
        'deleteLeasesForDevice' => {
            'ha_mode_primary' => 1,
            'params' => {
                name      => 'deleteLeasesForDevice',
                version   => $version,
                help      => 'triggers a delete of all lease entries for a mac address',
                code      => \&xmlrpc_deleteLeasesForDevice,
                signature => ['struct struct']
            }
        }
    };

    # we are still in setup_methods and will configure the methods from above hash
    my $ha_mode = $self->getHaMode();
    foreach my $method_name ( keys %{$methods} ) {
        my %params            = %{ $methods->{$method_name}->{'params'} };
        my $configured_method = $self->get_method($method_name);
        if ( $methods->{$method_name}->{$ha_mode} ) {
            if ( !ref $configured_method ) {
                $self->add_method( \%params );
            }
        } else {
            if ( ref $configured_method ) {
                $self->delete_method( $params{'name'} );
            }
        }
    }
    return $self;
}

package main;
use strict;
use warnings;

my $sc_controld = ScControld->new();

# add xmlrpc methods
$sc_controld->setup_methods();

$sc_controld->server_loop(
    no_http => 'true',

    #server_type => 'Single',
    server_type => 'PreFork',

    #server_type => 'Fork',
    min_servers => '5',
    max_servers => '80',

    # min_spare_servers => '2',
    # max_spare_servers => '50',
    log_level           => '2',
    log_file            => 'Sys::Syslog',
    syslog_ident        => 'sc_controld',
    child_communication => '1'
);

# TODO: should this be deleted? Or moved to documentation somewhere?
# sub run_worker
# {
#     my $j = 0;
#     while (1) {
#         sleep(1);
#         print STDERR ("I am the working  worker (PID $$) in iteration # $j\n");
#         $j++;
#     }
# }
#
# sub fork_worker
# {
#     my $pid = fork;
#
#     ### trouble
#     if ( not defined $pid ) {
#
#         #      $self->log(1,"Bad fork [$!]");
#         sleep(5);
#
#         ### parent:
#     } elsif ($pid) {
#         print STDERR "Parent is going on....\n";
#
#         #$self->log(1,"Still in Parent");
#         ### child
#     } else {
#         $self->run_worker;
#         exit;
#     }
# }
#
# # overridden methods of Net::Server / RPC::XML::Server
# sub run_n_children_hook
# {
#     $self->log( 1, "Parent $$ run_n_childre_hook called\n" );
# }
#
# sub post_accept_hook
# {
# }
#
# sub child_init_hook
# {
# }
#
# sub pre_server_close_hook
# {
# }

