#!/usr/bin/perl -w

use strict;
use File::Basename;
use Log::Log4perl;
use Net::LDAP qw(LDAP_NO_SUCH_OBJECT);
use SmartClient::LDAP;
use SmartClient::Config;
use SmartClient::IP;

#------------------------------------------------------------------------------
# Globale Variablen
#------------------------------------------------------------------------------

my $prgname = basename($0);
my ($logger, @nip_ignore_nets, $Ldap_Conn, @header);
my ($leases_file, $leases_file_cleared, $test_all_macs, %remove_from_leases, @dhcpd_conf_files);

#------------------------------------------------------------------------------
# Usage
#------------------------------------------------------------------------------

sub usage {

	print STDERR "Usage: $prgname <leases file of dhcpd> <name of cleared leases file> [ mac1 mac2 ... macn ]\n";
	print STDERR "\n";
	print STDERR "       if list of mac-addresses is empty, then all mac's, which are found in\n";
	print STDERR "       LDAP and conf-files of DHCPD and not within LEASES2LDAP_IGNORE_NETWORKS\n";
	print STDERR "       (see SmartClient::Leases2ldap) where removed. This could take a while!\n";
	exit 1;

}

#------------------------------------------------------------------------------
# Eintrag bereinigen
#------------------------------------------------------------------------------

sub clearDhcpdLease {

	my $cl_fh = shift;
	my @the_entry = @_;
	if (@the_entry) {
		my $clear = 0;
		my $the_mac;
		my $the_ip;
		my $ignore_ip = 0;
		foreach my $entry_line (@the_entry) {
			$the_ip = $1 if ($entry_line =~ /^lease\s+([\d.]+)\s+\{/);
			$the_mac = $1 if ($entry_line =~ /^\s+hardware ethernet ([0-9a-f]{2}:[0-9a-f]{2}:[0-9a-f]{2}:[0-9a-f]{2}:[0-9a-f]{2}:[0-9a-f]{2});/);
		}
		foreach my $switch_net (@nip_ignore_nets) {
			if ($switch_net->overlaps($the_ip)) {
				$ignore_ip = 1;
				$logger->info("IP $the_ip will be ignored as configured by LEASES2LDAP_IGNORE_NETWORKS");
				last;
			}
		}
		if ($the_mac and $the_ip and ! $ignore_ip) {
			if ($test_all_macs and not defined($remove_from_leases{$the_mac})) {
				$remove_from_leases{$the_mac} = 0;
				my $found_mac_in_ldap = 0;
				my $found_mac_in_dhcpd_conf_files = 0;
				my $ldap_search = $Ldap_Conn->search(filter => "(macAddress=$the_mac)");
				if ($ldap_search->code and $ldap_search->code != LDAP_NO_SUCH_OBJECT) {
					$logger->error("LDAP-Error: " . $ldap_search->error);
				}
				elsif ($ldap_search->count > 0) {
					$found_mac_in_ldap = 1
				}
				if ($found_mac_in_ldap) {
					$logger->error("found $the_mac more than once in LDAP") if ($ldap_search->count > 1);
					my $ldap_entry = $ldap_search->entry(0);
					my @ldap_entry_oc = $ldap_entry->get_value('objectClass');
					if (grep /^scSoPc$/i, @ldap_entry_oc) {
						$found_mac_in_dhcpd_conf_files = 1;
					}
					else {
						foreach my $dhcpd_conf_file (@dhcpd_conf_files) {
							if (open(DHCPD_CONF, '<', $dhcpd_conf_file)) {
								while(defined(my $dhcpd_conf_line = <DHCPD_CONF>)) {
									if ($dhcpd_conf_line =~ /^\s+hardware ethernet $the_mac;/) {
										$logger->error("found $the_mac more than once in configuration files of dhcpd") if ($found_mac_in_dhcpd_conf_files);
										$found_mac_in_dhcpd_conf_files = 1;
									}
								}
								close(DHCPD_CONF);
							}
						}
					}
				}
				$remove_from_leases{$the_mac} = 1 if ($found_mac_in_dhcpd_conf_files and $found_mac_in_ldap)
			}
			$clear = 1 if (defined($remove_from_leases{$the_mac}) and $remove_from_leases{$the_mac});
		}
		if ($clear) {
			$logger->info("removing entry with MAC " . $the_mac . " and IP " . $the_ip . " from " . $leases_file);
		}
		else {
			print $cl_fh "$_\n" foreach (@the_entry);
		}
	}
	else {
		print $cl_fh "$_\n" foreach (@header);
	}

}

#------------------------------------------------------------------------------
# MAIN
#------------------------------------------------------------------------------

my $config = SmartClient::Config->new(
	"/etc/smartclient/base",
	"/etc/smartclient/ldap2dnsdhcpd",
	"/etc/smartclient/ldap2dhcp",
    "/etc/smartclient/leases2ldap"
);

$config->require(
	'LDAP2DNSDHCPD_LOG4PERL_CONF',
	'LDAP2DNSDHCPD_LDAP_ADMIN_DN',
	'LDAP2DNSDHCPD_LDAP_ADMIN_PW',
	'LDAP2DHCP_DHCPCONFPATH'
);

# init Logging
Log::Log4perl::init( $config->{'LDAP2DNSDHCPD_LOG4PERL_CONF'} );
$logger = Log::Log4perl->get_logger();

$logger->info( basename($0) . " running");

# LDAP-Connection
$Ldap_Conn = SmartClient::LDAP->new(
	'CONFIG'     => $config,
	'LDAPBINDDN' => $config->get('LDAP2DNSDHCPD_LDAP_ADMIN_DN'),
	'LDAPBINDPW' => $config->get('LDAP2DNSDHCPD_LDAP_ADMIN_PW'),
	'NEEDCAP'    => 'ro'
);

usage unless ($#ARGV >= 1);
$leases_file = shift(@ARGV);
$leases_file_cleared = shift(@ARGV);
$test_all_macs = 1;
foreach my $mac_to_removed (@ARGV) {
	$test_all_macs = 0;
	$remove_from_leases{$mac_to_removed} = 1;
}
@dhcpd_conf_files = glob($config->get('LDAP2DHCP_DHCPCONFPATH') . "/*.conf");

if ($config->{'LEASES2LDAP_IGNORE_NETWORKS'}) {
	my @ignore_nets = split(/\s+/, $config->{'LEASES2LDAP_IGNORE_NETWORKS'});
	foreach my $ignore_net (@ignore_nets) {
		my $nip_ignore_net = SmartClient::IP->new($ignore_net);
		if (ref $nip_ignore_net) {
			$logger->info("ignoring IP-Addresses from "
			. $nip_ignore_net->{'first_ip'} . " to " . $nip_ignore_net->{'last_ip'});
			push(@nip_ignore_nets, $nip_ignore_net) if (ref $nip_ignore_net);
		}
	}
}

open(LF, '<', $leases_file) or $logger->error("Error opening $leases_file for reading, Exiting") && die;
open(CLF, '>', $leases_file_cleared) or $logger->error("Error opening $leases_file_cleared for writing, Exiting") && die;

my $header_idx = -1;
my @akt_entry = ();
my $entry_idx = -1;
my $leases_line;
while (defined($leases_line = <LF>)) {
	chomp $leases_line;
	if ($leases_line !~ /^\s*(?:#.*|)$/) {
		if ($leases_line =~ /^lease/) {
			clearDhcpdLease(\*CLF, @akt_entry);
			@akt_entry = ();
			$entry_idx = -1;
		}
		$akt_entry[++$entry_idx] = $leases_line;
	}
	$header[++$header_idx] = $leases_line if (! @akt_entry);
}
clearDhcpdLease(\*CLF, @akt_entry);

close(CLF);
close(LF);

__END__

=pod

=head1 NAME

sc_cleardhcpdleases.pl

=head2 SVN Version

$Id: sc_cleardhcpdleases.pl 5787 2008-09-03 15:33:31Z joergs $

=head1 SYNOPSIS

sc_cleardhcpdleases.pl

=head1 DESCRIPTION

Programm to delete Entries from leases file of dhcpd which are already in LDAP and
configuration files of dhcpd. Entries within the LEASES2LDAP_IGNORE_NETWORKS in no
case are deleted (see B<SmartClient::Leases2ldap>). The cleared leases file is
named <name_of_leases_file>.cleared.

=head1 AUTHOR

Holger Zuchel <Holger.Zuchel@debeka.de>

=cut
